#ifndef __SEZNAM_IMPL_H__
#define __SEZNAM_IMPL_H__

#include <initializer_list>
#include <iostream>
#include <memory>
#include <utility>
#include <stdexcept>

template<class T, template<class> class Aloktor = std::allocator>
class seznam_impl
{
	struct prvek;
public:
	// Zveejovan datov typy
	using typ_hodnot = T;

	// Konstruktory a destruktor
	seznam_impl();
	seznam_impl(std::initializer_list<T> seznam_inicializtor);
	seznam_impl(const seznam_impl& si);
	seznam_impl(seznam_impl&& si);
	~seznam_impl();

	// Vkldn
	void vlo_na_konec(T& co);
	void vlo_na_konec(T&& co);

	// Odstraovn
	void odstra_prvn();
	void vyprzdni();

	// Pro ladn
	void vpis();

	// Prohozen dvou seznam
	void proho(seznam_impl<T, Aloktor>& s);
	void proho(seznam_impl<T, Aloktor>&& s);

	// Stav
	bool przdn() { return hlava == zarka; }

private:
	prvek* hlava;
	prvek* zarka;
};


template<class T, template<class T> class Aloktor>
struct seznam_impl<T, Aloktor>::prvek
{
	T data;
	prvek* dal;
	static Aloktor<prvek> aloktor;

	prvek(T& data, prvek* dal = nullptr) : data(data), dal(dal) {}
	prvek(T&& data, prvek* dal = nullptr) : data(std::move(data)), dal(dal) {}

	void* operator new(std::size_t)
	{
		return aloktor.allocate(1);
	}

	void operator delete(void* p)
	{
		aloktor.deallocate((prvek*)p, 1);
	}
};

// Statick datov sloka prvku (aloktor)
template<class T, template<class> class Aloktor>
Aloktor<typename seznam_impl<T, Aloktor>::prvek>
seznam_impl<T, Aloktor>::prvek::aloktor;

// Konstruktory

template<class T, template<class T> class Aloktor>
inline seznam_impl<T, Aloktor>::seznam_impl()
{
	hlava = zarka = new prvek(T{});
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl(std::initializer_list<T> seznam_inicializtor)
	:seznam_impl()
{
	for (T i : seznam_inicializtor)
	{
		vlo_na_konec(i);
	}
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl(const seznam_impl<T, Aloktor>& s)
	:seznam_impl()
{
	for (prvek* pom = s.hlava; pom != s.zarka; pom = pom->dal)
	{
		vlo_na_konec(pom->data);
	}
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::seznam_impl(seznam_impl<T, Aloktor>&& s)
	:seznam_impl()
{
	proho(s);
}

template<class T, template<class T> class Aloktor>
seznam_impl<T, Aloktor>::~seznam_impl()
{
	vyprzdni();
	delete hlava;
	//std::cout << "destrukce" << std::endl;
}


// Vkldn

template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::vlo_na_konec(T& co)
{
	zarka->data = co;
	zarka->dal = new prvek(T{});
	zarka = zarka->dal;
}

template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::vlo_na_konec(T&& co)
{
	zarka->data = std::move(co);
	zarka->dal = new prvek(T{});
	zarka = zarka->dal;
}

// Pro ladn

template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::vpis()
{
	for (prvek* pom = hlava; pom != zarka; pom = pom->dal)
	{
		std::cout << pom->data << " ";
	}
	std::cout << std::endl;
}

// Prohozen
template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::proho(seznam_impl<T, Aloktor>& s)
{
	std::swap(hlava, s.hlava);
	std::swap(zarka, s.zarka);
}

template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::proho(seznam_impl<T, Aloktor>&& s)
{
	proho(s);
}

// Stav
template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::odstra_prvn()
{
	if (przdn())
	{
		throw std::out_of_range("Vbr z przdnho seznamu");
	}
	auto pom = hlava;
	hlava = pom->dal;
	delete pom;
}

template<class T, template<class T> class Aloktor>
void seznam_impl<T, Aloktor>::vyprzdni()
{
	while (!przdn())
	{
		odstra_prvn();
	}
}


#endif // __SEZNAM_IMPL_H__